extern crate sdl2;

use std::thread::sleep;
use std::time::{Duration, SystemTime, SystemTimeError, UNIX_EPOCH};

use bitvec::prelude::*;
use sdl2::Sdl;
use sdl2::event::Event;
use sdl2::keyboard::Keycode;
use sdl2::pixels::{Color, PixelFormatEnum};
use sdl2::rect::Rect;
use sdl2::render::{WindowCanvas, Texture, TextureCreator};
use sdl2::video::{Window, WindowContext};

pub mod cp437;
pub mod ega;
use cp437::GLYPHS;
use ega::EgaColors;


const SCREEN_HEIGHT: u32 = 350;
const SCREEN_WIDTH: u32 = 640;

const GLYPH_COLUMNS: u8 = 80;
const GLYPH_ROWS: u8 = 25;

const GLYPH_HEIGHT: u8 = 14;
const GLYPH_WIDTH: u8 = 8;

const GLYPH_ATLAS_HEIGHT: usize = GLYPH_HEIGHT as usize * 8;
const GLYPH_ATLAS_WIDTH: usize = GLYPH_WIDTH as usize * 32;

// Total glyph textures = 256 glyphs * 16 foreground colors * 8 background colors
const TOTAL_GLYPH_ATLASES: u8 = 16 * 8;

// Total screen blocks = GLYPH_COLUMNS * GLYPH_ROWS
const TOTAL_SCREEN_BLOCKS: usize = 2000;

const FPS_60_HERTZ: u32 = 1_000_000_000u32 / 60;
const BLINK_RATE: u128 = 266666666;

#[derive(Clone, Copy)]
pub struct ScreenBlock {
    pub blink: bool,
    pub background: u8,
    pub foreground: u8,
    pub glyph: u8,
}

impl ScreenBlock {
    pub fn new(blink: bool, background: u8, foreground: u8, glyph: u8) -> Self {
        ScreenBlock { blink, background, foreground, glyph }
    }

    pub fn as_index(&self) -> usize {
        EgaColors::textattr_to_u8(false, self.background, self.foreground) as usize
    }

    pub fn to_canvas(
        &self,
        canvas: &mut WindowCanvas,
        textures: &Vec<Texture>,
        x: i32,
        y: i32,
        blink_state: &bool
    ) {
    
        let tex_idx = match blink_state {
            false => self.as_index(),
            true => match self.blink {
                false => self.as_index(),
                true => EgaColors::textattr_to_u8(false, self.background, self.background) as usize,
            },
        };
    
        let src_x = (self.glyph % 32 * GLYPH_WIDTH) as i32;
        let src_y = (self.glyph / 32 * GLYPH_HEIGHT) as i32;
        
        // These next two assignments look odd mainly because of Rust's current
        // issue with exclusive range patterns. See here:
        // https://github.com/rust-lang/rust/issues/37854
    
        let dest_x = match x {
            n if (0..=GLYPH_COLUMNS as i32 - 1).contains(&n) => GLYPH_WIDTH as i32 * x,
            _ => panic!("x is out of range! Must be 0..79 inclusive."),
        };
    
        let dest_y = match y {
            n if (0..=GLYPH_ROWS as i32 - 1).contains(&n) => GLYPH_HEIGHT as i32 * y,
            _ => panic!("y is out of range! Must be 0..24 inclusive."),
        };
    
        match canvas.copy(
            &textures[tex_idx],
            Rect::new(src_x, src_y, u32::from(GLYPH_WIDTH), u32::from(GLYPH_HEIGHT)),
            Rect::new(dest_x, dest_y, u32::from(GLYPH_WIDTH), u32::from(GLYPH_HEIGHT)),
        ) {
            Ok(()) => (),
            Err(err) => panic!("Unable to copy texture to canvas! SDL Error: {}", err),
        };
    }
}

pub fn init_sdl() -> (Sdl, Window) {
    let sdl_context = match sdl2::init() {
        Ok(sdl_context) => sdl_context,
        Err(err) => panic!("SDL could not initialize! SDL_Error: {}", err),
    };

    let video_subsystem = match sdl_context.video() {
        Ok(video) => video,
        Err(err) => panic!("Could not obtain handle to the video subsystem! SDL_Error: {}", err),
    };

    let window = match video_subsystem.window("DOS Text Mode", SCREEN_WIDTH, SCREEN_HEIGHT)
        .position_centered()
        .opengl()
        .build() {
            Ok(window) => window,
            Err(err) => panic!("SDL could not create a window! SDL_Error: {}", err),
        };
    (sdl_context, window)
}

pub fn init_glyph_atlases(creator: &TextureCreator<WindowContext>) -> Vec<Texture<'_>> {
    let mut atlases: Vec<Texture> = Vec::new();
    for atlas_idx in 0..TOTAL_GLYPH_ATLASES {
        let (_, back, fore) = EgaColors::u8_to_textattr(atlas_idx.try_into().unwrap());
        let mut texture = match creator.create_texture_streaming(PixelFormatEnum::RGB24, GLYPH_ATLAS_WIDTH as u32, GLYPH_ATLAS_HEIGHT as u32) {
            Ok(texture) => texture,
            Err(err) => panic!("SDL could not create texture. SDL Error: {}", err),
        };
        match texture.with_lock(None, |buffer: &mut [u8], pitch: usize| {
            for (glyph_num, glyph_data) in GLYPHS.iter().enumerate() {
                let glyph_x = glyph_num % 32;
                let glyph_y = glyph_num / 32;
                for (local_y, line) in glyph_data.iter().enumerate() {
                    let bits = line.view_bits::<Msb0>();
                    for (local_x, pixel) in bits.iter().enumerate() {
                        let offset = ((glyph_y * GLYPH_HEIGHT as usize * pitch) + (local_y * pitch)) + ((glyph_x * GLYPH_WIDTH as usize) + local_x) * 3;
                        if pixel == false {
                            buffer[offset] = EgaColors::u8_to_rgb(back).r;
                            buffer[offset + 1] = EgaColors::u8_to_rgb(back).g;
                            buffer[offset + 2] = EgaColors::u8_to_rgb(back).b;
                        } else {
                            buffer[offset] = EgaColors::u8_to_rgb(fore).r;
                            buffer[offset + 1] = EgaColors::u8_to_rgb(fore).g;
                            buffer[offset + 2] = EgaColors::u8_to_rgb(fore).b;
                        }
                    }
                }
            }
        }) {
            Ok(()) => (),
            Err(err) => panic!("Unable to modify texture. SDL Error: {}", err),
        };
        atlases.push(texture);
    }
    atlases
}

pub fn create_test_screen(screen: &mut [ScreenBlock]) {
    for col in 0..60 {
        screen[col as usize] = ScreenBlock::new(false, EgaColors::Black.into(), EgaColors::BrightYellow.into(), 178);
    }
    for row in 1..GLYPH_ROWS-1 {
        let idx = row as usize * GLYPH_COLUMNS as usize;
        screen[idx] = ScreenBlock::new(false, EgaColors::Black.into(), EgaColors::BrightYellow.into(), 178);
        screen[idx + 59] = ScreenBlock::new(false, EgaColors::Black.into(), EgaColors::BrightYellow.into(), 178);
    }
    for col in 0..60 {
        let idx = (GLYPH_ROWS - 1) as usize * GLYPH_COLUMNS as usize + col as usize;
        screen[idx] = ScreenBlock::new(false, EgaColors::Black.into(), EgaColors::BrightYellow.into(), 178);
    }

    screen[1030] = ScreenBlock::new(true, EgaColors::Blue.into(), EgaColors::LightGray.into(), 178);
}

pub fn update(elapsed: u128) {
    //println!("Updated after {} nanoseconds.", elapsed);
}

pub fn render(canvas: &mut WindowCanvas, atlases: &Vec<Texture>, screen: &[ScreenBlock], blink: &bool) {
    canvas.set_draw_color::<Color>(EgaColors::Black.into());
    canvas.clear();

    for (idx, block) in screen.iter().enumerate() {
        let x = (idx % GLYPH_COLUMNS as usize) as i32;
        let y = (idx / GLYPH_COLUMNS as usize) as i32;
        block.to_canvas(canvas, &atlases, x, y, blink);
    }

    canvas.present();
}

pub fn get_timestamp_in_nanos() -> Result<u128, SystemTimeError> {
    let current_systime = SystemTime::now();
    let since_epoch = current_systime.duration_since(UNIX_EPOCH)?;
    let nanos = since_epoch.as_nanos();

    Ok(nanos)
}

pub fn main() {
    let (sdl_context, window) = init_sdl();

    let mut canvas = match window.into_canvas().build() {
        Ok(canvas) => canvas,
        Err(err) => panic!("SDL could not create a canvas! SDL_Error: {}", err),
    };
    
    let texture_creator = canvas.texture_creator();
    let atlases: Vec<Texture> = init_glyph_atlases(&texture_creator);
    
    let mut dos_screen: [ScreenBlock; TOTAL_SCREEN_BLOCKS] = [ScreenBlock::new(false, EgaColors::Black.into(), EgaColors::Black.into(), 0); TOTAL_SCREEN_BLOCKS];

    create_test_screen(&mut dos_screen);

    let mut blink_tick: u128 = 0;
    let mut blink_state: bool = false;

    render(&mut canvas, &atlases, &dos_screen, &blink_state);

    let mut event_pump = match sdl_context.event_pump() {
        Ok(event_pump) => event_pump,
        Err(err) => panic!("Could not obtain SDL2 Event Pump! SDL_Error: {}", err),
    };

    let mut prev_tick = match get_timestamp_in_nanos() {
        Ok(prev_tick) => prev_tick,
        Err(err) => panic!("Could not get the current timestamp: {}", err),
    };

    'running: loop {
        let curr_tick = match get_timestamp_in_nanos() {
            Ok(curr_tick) => curr_tick,
            Err(err) => panic!("Could not get the current timestamp: {}", err),
        };
        let elapsed = curr_tick - prev_tick;

        for event in event_pump.poll_iter() {
            match event {
                Event::Quit {..} |
                Event::KeyDown { keycode: Some(Keycode::Escape), .. } => {
                    break 'running
                },
                _ => {}
            }
        }

        update(elapsed);

        if blink_tick >= BLINK_RATE {
            blink_state = !blink_state;
            blink_tick = 0;
        } else {
            blink_tick += elapsed;
        }

        render(&mut canvas, &atlases, &dos_screen, &blink_state);
        //sleep(Duration::new(0, FPS_60_HERTZ));

        prev_tick = curr_tick;
    }
}
